@@ -47,23 +47,68 @@ u_int gFontsHeight = 0;
 u_char sgFontsDefaultGlyph[] = {0x80};
 u_char sgFontsDefaultDGlyph[] = {0xC0};
 
-#define FMACRO_94__FONT(final, aline, name) {\
-	tfont_default_glyph, name, 1, 1,\
-	TFONT_FT_SINGLE | TFONT_FT_94CHAR| final,\
-	aline, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,\
-	0xFF, 0x00, 0xFF, 0x00, 0x00, 2, 1 }
+#define FMACRO_94__FONT(final, aline, fontname) {\
+	.conv = tfont_default_glyph, \
+	.name = fontname, \
+	.width = 1, \
+	.height = 1, \
+	.fsignature = TFONT_FT_SINGLE | TFONT_FT_94CHAR| final,\
+	.fhalf = aline, \
+	.aliasF = TFONT_ALIAS, \
+	.glyph = NULL, \
+	.glyph_width = NULL, \
+	.dglyph = sgFontsDefaultGlyph, \
+	.bitmap = NULL,\
+	.colf = 0xFF, \
+	.coll = 0x00, \
+	.rowf = 0xFF, \
+	.rowl = 0x00, \
+	.colspan = 0x00, \
+	.bytew = 2, \
+	.bytec = 1 \
+}
 
-#define FMACRO_96__FONT(final, aline, name) {\
-	tfont_default_glyph, name, 1, 1,\
-	TFONT_FT_SINGLE | TFONT_FT_96CHAR| final,\
-	aline, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,\
-	0xFF, 0x00, 0xFF, 0x00, 0x00, 2, 1 }
+#define FMACRO_96__FONT(final, aline, fontname) {\
+	.conv = tfont_default_glyph, \
+	.name = fontname, \
+	.width = 1, \
+	.height = 1, \
+	.fsignature = TFONT_FT_SINGLE | TFONT_FT_96CHAR| final,\
+	.fhalf = aline, \
+	.aliasF = TFONT_ALIAS, \
+	.glyph = NULL, \
+	.glyph_width = NULL, \
+	.dglyph = sgFontsDefaultGlyph, \
+	.bitmap = NULL,\
+	.colf = 0xFF, \
+	.coll = 0x00, \
+	.rowf = 0xFF, \
+	.rowl = 0x00, \
+	.colspan = 0x00, \
+	.bytew = 2, \
+	.bytec = 1 \
+}
 
-#define FMACRO_94N_FONT(final, aline, name) {\
-	tfont_default_glyph, name, 2, 1,\
-	TFONT_FT_DOUBLE | TFONT_FT_94CHAR| final,\
-	aline, TFONT_ALIAS, NULL, sgFontsDefaultDGlyph, NULL,\
-	0xFF, 0x00, 0xFF, 0x00, 0x00, 2, 2 }
+#define FMACRO_94N_FONT(final, aline, fontname) {\
+	.conv = tfont_default_glyph, \
+	.name = fontname, \
+	.width = 2, \
+	.height = 1, \
+	.fsignature = TFONT_FT_DOUBLE | TFONT_FT_94CHAR| final,\
+	.fhalf = aline, \
+	.aliasF = TFONT_ALIAS, \
+	.glyph = NULL, \
+	.glyph_width = NULL, \
+	.dglyph = sgFontsDefaultDGlyph, \
+	.bitmap = NULL,\
+	.colf = 0xFF, \
+	.coll = 0x00, \
+	.rowf = 0xFF, \
+	.rowl = 0x00, \
+	.colspan = 0x00, \
+	.bytew = 2, \
+	.bytec = 2 \
+}
 
 TFont gFont[] = {
 	FMACRO_94__FONT(0x40, FH_LEFT , "iso646-1973irv"),
@@ -159,144 +204,47 @@ TFont gFont[] = {
 	FMACRO_94__FONT(0x64, FH_LEFT , "ccitt-mosaic-sup3"),
 	/* Japano */
 	FMACRO_94__FONT(0x49, FH_LEFT , "jisc6220-1969kana"),
-#if 0
-					/* 1st element shall be ASCII */
-	{ tfont_default_glyph,	"ASCII", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_94CHAR| 0x42,
-		FH_LEFT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-					/* 2nd element shall be JIS X 0208 */
-	{ tfont_default_glyph,	"JISX0208-1978", 2, 1,
-		TFONT_FT_DOUBLE | TFONT_FT_94CHAR | 0x40,
-		FH_LEFT, TFONT_ALIAS, NULL, sgFontsDefaultDGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"JISX0208-1983", 2, 1,
-		TFONT_FT_DOUBLE | TFONT_FT_94CHAR | 0x42,
-		FH_LEFT, TFONT_ALIAS, NULL, sgFontsDefaultDGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"CODEPAGE-437", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 'U',
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ALT-ASCII", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_94CHAR| '0',
-		FH_LEFT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"JISX0201-ROMAN", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_94CHAR| 0x4A,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"JISX0201-KATAKANA", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_94CHAR| 0x49,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"GB2312", 2, 1,
-		TFONT_FT_DOUBLE | TFONT_FT_94CHAR | 0x41,
-		FH_LEFT, TFONT_ALIAS, NULL, sgFontsDefaultDGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"KSX1001", 2, 1,
-		TFONT_FT_DOUBLE | TFONT_FT_94CHAR | 0x43,
-		FH_LEFT, TFONT_ALIAS, NULL, sgFontsDefaultDGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-1", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x41,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-2", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x42,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-3", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x43,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-4", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x44,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-5", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x4C,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-6", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x41,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-7", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x46,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-8", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x48,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-9", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x4D,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-10", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x56,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-11", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x54,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-13", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x59,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-14", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x5f,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-15", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x62,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ISO8859-16", 1, 1,
-		TFONT_FT_SINGLE|TFONT_FT_96CHAR| 0x66,
-		FH_RIGHT, TFONT_ALIAS, NULL, sgFontsDefaultGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"JISX0212", 2, 1,
-		TFONT_FT_DOUBLE | TFONT_FT_94CHAR | 0x44,
-		FH_LEFT, TFONT_ALIAS, NULL, sgFontsDefaultDGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
-	},
-	{ tfont_default_glyph,	"ITU-TT.101ChinesePrimarySet", 2, 1,
-		TFONT_FT_DOUBLE | TFONT_FT_94CHAR | 0x45,
-		FH_LEFT, TFONT_ALIAS, NULL, sgFontsDefaultDGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
+#ifdef JFB_UTF8
+	{
+	    .conv = tfont_default_glyph,
+	    .name = "iso10646.1",  /* Unicode 2.0 */
+	    .width = 1, 
+	    .height = 1,
+	    .fsignature = TFONT_FT_OTHER,
+	    .fhalf = FH_UNI, 
+	    .aliasF = TFONT_ALIAS, 
+	    .glyph = NULL, 
+	    .glyph_width = NULL,
+	    .dglyph = sgFontsDefaultDGlyph, 
+	    .bitmap = NULL,
+	    .colf = 0xFF, 
+	    .coll = 0x00, 
+	    .rowf = 0xFF, 
+	    .rowl = 0x00, 
+	    .colspan = 0x00, 
+	    .bytew = 1, 
+	    .bytec = 1
 	},
 #endif
-	{	tfont_default_glyph,	NULL, 1, 1,
-		0x00000000,
-		FH_LEFT, TFONT_ALIAS, NULL, sgFontsDefaultDGlyph, NULL,
-		0xFF, 0x00, 0xFF, 0x00, 0x00, 1, 1
+	{
+	    .conv = tfont_default_glyph,
+	    .name = NULL, 
+	    .width = 1, 
+	    .height = 1,
+	    .fsignature = 0x00000000,
+	    .fhalf = FH_LEFT, 
+	    .aliasF = TFONT_ALIAS, 
+	    .glyph = NULL, 
+	    .glyph_width = NULL,
+	    .dglyph = sgFontsDefaultDGlyph, 
+	    .bitmap = NULL,
+	    .colf = 0xFF, 
+	    .coll = 0x00, 
+	    .rowf = 0xFF, 
+	    .rowl = 0x00, 
+	    .colspan = 0x00, 
+	    .bytew = 1, 
+	    .bytec = 1
 	},
 };
 
@@ -309,6 +257,7 @@ void tfont_ary_show_list(FILE* fp)
 	u_int d;
 	u_int f1;
 	u_int f2;
+	u_int uni;
 	while (ce->name) {
 		n = ce->fsignature;
 		c = n & 255;
@@ -316,11 +265,16 @@ void tfont_ary_show_list(FILE* fp)
 		d = (n >> 0) & 15;
 		f1 = n & TFONT_FT_DOUBLE;
 		f2 = n & TFONT_FT_96CHAR;
-		fprintf(fp, " [%02u/%02u(%c) %s%s] %s\n",
-			u, d, c,
-			f2 ? "96" : "94",
-			f1 ? "^N" : "  ",
-			ce->name);
+		uni = ce->fhalf == FH_UNI ? 1 : 0;
+		if (uni) {
+		    fprintf(fp, " [--/--(-) uni ] %s\n", ce->name);
+		} else {
+		    fprintf(fp, " [%02u/%02u(%c) %s%s] %s\n",
+			    u, d, c,
+			    f2 ? "96" : "94",
+			    f1 ? "^N" : "  ",
+			    ce->name);
+		}
 		ce++;
 	}
 	
@@ -369,7 +323,8 @@ void tfont_final(TFont* p)
 
 const u_char* tfont_default_glyph(
 	TFont* p,
-	u_int chlw)
+	u_int chlw,
+	u_int* width)
 {
 	static u_char gly[PICOFONT_HEIGHT * 2];
 
@@ -377,6 +332,7 @@ const u_char* tfont_default_glyph(
 	u_int i;
 	u_int a;
 	u_int b;
+	*width = p->width;
 
 	*cp++ = 0x80;
 	*cp++ = 0x00;
@@ -407,16 +363,20 @@ const u_char* tfont_default_glyph(
 
 const u_char* tfont_standard_glyph(
 	TFont* p,
-	u_int chlw)
+	u_int chlw,
+	u_int *width)
 {
 	u_char b2 = (chlw >> 8) & 0xFF;
 	u_char b1 = chlw & 0xFF;
 	u_int i;
 
+	*width = p->width;
 	if (b1 < p->colf || p->coll < b1 || b2 < p->rowf || p->rowl < b2) {
 		return p->dglyph;	
 	} else {
 		i = (b1 - p->colf) + (b2  - p->rowf) * p->colspan;
+		if (p->glyph_width)
+			*width = p->glyph_width[i];
 		return p->glyph[i];
 	}
 }
@@ -468,7 +428,6 @@ static void tfont_setup_font(TFont* pf, const char* fname, FONTSET_HALF hf)
 	tpcf_as_tfont(&pcf, pf);
 	pf->fhalf = hf;
 	pf->conv = tfont_standard_glyph;
-
 	tpcf_final(&pcf);
 }
 
@@ -480,6 +439,7 @@ static void tfont_alias(TFont* dst, TFont* src, FONTSET_HALF hf)
 	dst->fhalf = hf;
 	dst->aliasF = TFONT_ALIAS;
 	dst->glyph = src->glyph;
+	dst->glyph_width = src->glyph_width;
 	dst->dglyph = src->dglyph;
 	dst->bitmap = src->bitmap;
 	dst->colf = src->colf; 
@@ -528,10 +488,15 @@ int tfont_is_valid(TFont* p)
 	return (p->width == 0 || p->height == 0) ? 0 : 1;
 }
 
+int tfont_is_loaded(TFont *p)
+{
+	return (p->conv != tfont_default_glyph);
+}
+
 void tfont_setup_fontlist(TCapValue* vp)
 {
 	static const char* types[] = {"pcf", "alias", NULL}; 
-	static const char* sides[] = {"L", "R", NULL}; 
+	static const char* sides[] = {"L", "R", "U", NULL}; 
 
 	const char* srn;
 	const char* sty;
@@ -562,7 +527,6 @@ void tfont_setup_fontlist(TCapValue* vp)
 
 		type = util_search_string(sty, types);
 		side = util_search_string(shf, sides);
-
 		if (type == -1 || side == -1) {
 			print_message("FONT : Skipped (BAD FORMAT)\n");
 			goto FINALIZE;
@@ -572,18 +536,22 @@ void tfont_setup_fontlist(TCapValue* vp)
 			print_message("FONT : Skipped (Unknown FONTSET=`%s'.)\n", srn);
 			goto FINALIZE;
 		}
+		switch (side) {
+		case 0: side = FH_LEFT; break;
+		case 1: side = FH_RIGHT; break;
+		case 2: side = FH_UNI; break;
+		}
+
 		switch (type) {
 		case 0:		/* pcf file */
-			tfont_setup_font(dst, sph,
-					 (side == 0) ? FH_LEFT : FH_RIGHT);
+			tfont_setup_font(dst, sph, side);
 			break;
 		case 1:		/* alias */
 			if (!(src = tfont_ary_search(sph))) {
 				print_message("FONT : Skipped (Unknown ALIAS FONTSET=`%s'.)\n", sph);
 				continue;
 			}
-			tfont_alias(dst, src,
-					(side == 0) ? FH_LEFT : FH_RIGHT);
+			tfont_alias(dst, src, side);
 			break;
 		}
 
@@ -593,70 +561,22 @@ void tfont_setup_fontlist(TCapValue* vp)
 
 	tfont_fontlist_glyph_size();
 
-	if (!tfont_is_valid(&(gFont[0]))) {
-		die("FONT : ISO8859 not loaded.\n");
-	}
-}
-
-void tfont_show_sequence(FILE *tf, const char *en)
-{
-	static const char *table[] = {"G0", "G1", "G2", "G3", NULL};
-	static const char *invokeL[] = {"\017", "\016", "\033n", "\033o"};
-	static const char *invokeR[] = {"", "\033~", "\033}", "\033|"};
-	static const char csr4[] = {ISO_GZD4, ISO_G1D4, ISO_G2D4, ISO_G3D4};
-	static const char csr6[] = {MULE__GZD6, ISO_G1D6, ISO_G2D6, ISO_G3D6};
-	TCsv farg;
-	int ig;
-	int i;
-	const char *g;
-
-	tcsv_init(&farg, en);
-	if (farg.cap != 6) {
-		print_warn("ENCODING : BAD FORMAT : %s\n", en);
-		goto end;
-	}
-	/* ISO-2022 */
-	printf("%s", "\033%@");
-	
-	/* GL */
-	g = tcsv_get_token(&farg);
-	ig = util_search_string(g, table);
-	if (ig == -1) {
-		print_error("ENCODING : GL %s ? (BAD FORMAT)\n", g);
-		goto end;
-	}
-	fprintf(tf, "%s", invokeL[ig]);
-
-	/* GR */
-	g = tcsv_get_token(&farg);
-	ig = util_search_string(g, table);
-	if (ig == -1) {
-		print_error("ENCODING : GR %s ? (BAD FORMAT)\n", g);
-		goto end;
-	}
-	fprintf(tf, "%s", invokeR[ig]);
-
-	/* G0 .. G3 */
-	for (i = 0; i < 4; i++) {
-		TFont *f;
-		u_int n;
-		char c;
-		u_int f1, f2;
-		g = tcsv_get_token(&farg);
-		f = tfont_ary_search(g);
-		if (f == NULL) {
-			print_error("ENCODING : G%d %s ? (BAD FORMAT)\n", i, g);
-			continue;
+	if (tfont_is_loaded(&(gFont[0])) == 0) {
+#ifdef JFB_UTF8
+		int i, found = 0;
+		for (i = 1; gFont[i].fsignature; i++) {
+			if (gFont[i].fhalf == FH_UNI) {
+				if (tfont_is_loaded(&(gFont[i]))) {
+					found = 1;
+				}
+				break;
+			}
 		}
-		n = f->fsignature;
-		c = n & 255;
-		f1 = n & TFONT_FT_DOUBLE;
-		f2 = n & TFONT_FT_96CHAR;
-		fprintf(tf, "\033%s%c%c", (f1 ? "$" : ""), 
-			(f2 ? csr6[i] : csr4[i]), c);
-	}
-end:
-	tcsv_final(&farg);
-	return;
-}
+		if (!found)
+			die("FONT : ISO8859/ISO10646 not loaded.\n");
 
+#else
+		die("FONT : ISO8859 not loaded.\n");
+#endif
+	}
+}
