@@ -41,6 +41,7 @@
 #include <signal.h>
 #include <fcntl.h>
 #include <sys/stat.h>
+#include <sys/wait.h>
 #include <getopt.h>
 #include <locale.h>
 #include <langinfo.h>
@@ -54,9 +55,9 @@
 #include "mytypes.h"
 #include "getcap.h"
 #include "util.h"
+#include "csv.h"
 #include "main.h"
 
-
 #ifdef SYSCONFDIR
 #define	FILE_JFBTERM_CONF	SYSCONFDIR"/jfbterm.conf"
 #else
@@ -131,15 +132,20 @@ void tapp_change_to_original_console(TApplication* p)
         signal(SIGUSR2, SIG_DFL);
 */
 
-        cfd = util_privilege_open("/dev/console", O_WRONLY);
-        if (cfd < 0 && 
-	    (cfd = util_privilege_open("/dev/console", O_RDONLY)) < 0) {
+	util_privilege_on();
+	cfd = open("/dev/console", O_WRONLY);
+	if (cfd < 0 && 
+	    (cfd = open("/dev/console", O_RDONLY)) < 0) {
+		util_privilege_off();
                 print_strerror("/dev/console");
-        }
-        if (ioctl(cfd, VT_ACTIVATE, n) != 0) {
-                print_warn("can't activate VC(%d)", n);
-        }
-        close(cfd);
+		return;
+	}
+	if (ioctl(cfd, VT_ACTIVATE, n) != 0) {
+		util_privilege_off();
+		print_warn("can't activate VC(%d)", n);
+	}
+	util_privilege_off();
+	close(cfd);
 }
 
 void tapp_final(TApplication* p)
@@ -147,9 +153,6 @@ void tapp_final(TApplication* p)
 	if (p->gCapsQ) {
 		tcaps_final(&(p->gCaps));
 	}
-	if (p->gOrigVirtualConsoleQ) {
-		tapp_change_to_original_console(p);
-	}
 	util_free(p->gExecShellArgv);
 }
 
@@ -164,7 +167,6 @@ void tapp_init(TApplication* p)
 {
 	static char shell[128];
 	
-	p->gOrigVirtualConsoleQ	= FALSE;
 	p->gOrigVirtualConsole	= -1;
 	tcaps_init(&(p->gCaps));
 	p->gCapsQ	= TRUE;
@@ -209,7 +211,6 @@ void tapp_change_to_new_console(TApplication* p)
 #endif
         ioctl(cfd, VT_GETSTATE, &vts);
 	p->gOrigVirtualConsole	= vts.v_active;
-	p->gOrigVirtualConsoleQ	= TRUE;
 
         ioctl(cfd, VT_OPENQRY, &vtNum);
         if (vtNum < 0) {
@@ -220,21 +221,26 @@ void tapp_change_to_new_console(TApplication* p)
                 print_strerror_and_exit("fork");
         }
         if (child) {
+		/* parent - waiting child process, and reset VC */
                 print_message("\r\nJFBTERM> switched to new VC\r\n");
-		p->gOrigVirtualConsoleQ	= FALSE;
+	        if (waitpid(child, NULL, 0) < 0) {
+			print_strerror("waitpid");
+		}
+		tapp_change_to_original_console(p);
                 exit(EXIT_SUCCESS);
         }
         setsid();
         sprintf(vtty, "/dev/tty%d", vtNum);
-        if ((vfd = util_privilege_open(vtty, O_RDWR)) < 0) {
+	util_privilege_on();
+        if ((vfd = open(vtty, O_RDWR)) < 0) {
+		util_privilege_off();
                 die("can't open %s", vtty);
         }
         if (ioctl(cfd, VT_ACTIVATE, vtNum) != 0) {
+		util_privilege_off();
                 die("can't activate VC(%d)", vtNum);
         }
-/*
-        atexit(change_to_original_console);
-*/
+	util_privilege_off();
         close(cfd);
         dup2(vfd, 0);
         dup2(vfd, 1);
@@ -292,24 +298,26 @@ void ShowCaps(void)
 	fprintf(stdout, "[[ FONTSET ]]\n");
 	tfont_ary_show_list(stdout);
 
-	fprintf(stdout,
-		"[[ ENCODING ]]\n"
+	fprintf(stdout, "[[ ENCODING ]]\n");
 #ifdef JFB_MINI_JFBTERM
-			" ISO-2022-JP EUC-JP"
+	fprintf(stdout, " ISO-2022-JP EUC-JP");
 #else
-			" ISO-2022-*"
-#if defined(JFB_UNICODE) && defined(JFB_UTF_7)
-			" UTF-7"
-#endif	/* JFB_UNICODE && JFB_UTF_7 */
-#if defined(JFB_UNICODE) && defined(JFB_UTF_8)
-			" UTF-8"
-#endif	/* JFB_UNICODE && JFB_UTF_8 */
+	fprintf(stdout, " ISO-2022-*");
+#ifdef JFB_UTF7
+	fprintf(stdout, " UTF-7");
+#endif
+#ifdef JFB_UTF8
+	fprintf(stdout, " UTF-8");
+#endif
 #ifdef JFB_TRON
-			" TRON"
+	fprintf(stdout, " TRON");
 #endif	/* JFB_TRON */
+#ifdef JFB_OTHER_CODING_SYSTEM
+	fprintf(stdout, " OTHER-CODING-SYSTEM");
+#endif
 #endif	/* JFB_MINI_JFBTERM */
-		"\n"
-		"[[ MISC ]]\n"
+	fprintf(stdout, "\n");
+	fprintf(stdout, "[[ MISC ]]\n"
 #ifdef DEBUG
 			" DEBUG"
 #endif
@@ -333,20 +341,25 @@ tapp_setup_encoding(char *en)
 		print_message("ENCODING: locale = %s\n", en);
 	}
 	if (strchr(en, ',') == NULL) {
-		char *encname;
-
-		encname = (char *)malloc(strlen("encoding.") + strlen(en) + 1);
-		strcpy(encname, "encoding.");
-		strcat(encname, en);
-
-		en = tcaps_find_first(&(gApp.gCaps), encname);
+		char *encname = en;
+		en = tcaps_find_entry(&(gApp.gCaps), "encoding.", en);
 		if (!en)
 			print_error("%s not found, fallback to default\n", encname);
-		free(encname);
 	}
 	if (!en) {
 		en = "G0,G1,ansix3.4-1968,ansix3.4-1968,iso8859.1-1987,ansix3.4-1968";
 	}
+#ifdef JFB_OTHER_CODING_SYSTEM
+	{
+		/* XXX : initialize other-coding-system */
+		TCapability* fother;
+		fother = tcaps_find(&(gApp.gCaps), "other-coding-system");
+		if (fother && fother->values) {
+			tvterm_setup_otherCodingSystem(&(gApp.gCaps), 
+						       fother->values);
+		}
+	}
+#endif
 	return en;
 }
 
@@ -381,7 +394,7 @@ int main(int argc, char *argv[])
 		FILE *tf = fopen("/dev/tty", "w");
 		if (tf == NULL)
 			tf = stderr; /* XXX */
-		tfont_show_sequence(tf, en);
+		tvterm_show_sequence(tf, &(gApp.gCaps), en);
 		fclose(tf);
 		printf("ENCODING: %s\n", en);
 		exit(0);
@@ -402,20 +415,6 @@ int main(int argc, char *argv[])
 	if ((cp = tcaps_find_first(&(gApp.gCaps), "color.gamma")) != NULL) {
 		fbgamma = (float)atof(cp);
 	}
-
-	tapp_change_to_new_console(&gApp);
-
-	tfbm_init(&gFramebuffer);
-	
-	tfont_setup_fontlist(fcap->values);
-
-	tfbm_open(&gFramebuffer);
-
-	tn = tcaps_find_first(&(gApp.gCaps), "term");
-	if (!tn) {
-		tn = "jfbterm";
-	}
-
 	print_message("encoding : %s\n", en);
 	print_message("exec : %s ", gApp.gExecShell);
 	{
@@ -424,10 +423,19 @@ int main(int argc, char *argv[])
 			print_message("%s ", gApp.gExecShellArgv[i]);
 	}
 
+	tapp_change_to_new_console(&gApp);
+	tfbm_init(&gFramebuffer);
+	tfont_setup_fontlist(fcap->values);
+	tfbm_open(&gFramebuffer);
+
+	util_privilege_drop();
+	tn = tcaps_find_first(&(gApp.gCaps), "term");
+	if (!tn) {
+		tn = "jfbterm";
+	}
 	tterm_start(&gTerm, tn, en);
 
 	tfbm_close(&gFramebuffer);
 	tfont_ary_final();
-
 	exit(EXIT_SUCCESS);
 }
