@@ -92,8 +92,9 @@ void tterm_init(TTerm* p, const char* en)
 	p->name[0] = '\0';
 	tcgetattr(0, &(p->ttysave));
 	tvterm_init(&(p->vterm), p,
-			gFramebuffer.width/gFontsWidth,
-			gFramebuffer.height/gFontsHeight, en);
+		    gFramebuffer.width/gFontsWidth,
+		    gFramebuffer.height/gFontsHeight, 
+		    &(gApp.gCaps), en);
 }
 
 void tterm_final(TTerm* p)
@@ -163,7 +164,6 @@ void tterm_start(TTerm* p, const char* tn, const char* en)
 	gChildProcessId = fork();
 	if (gChildProcessId == 0) {
 	    /* child */
-	    util_privilege_drop();
 	    tterm_wakeup_shell(p, tn);
 	    exit(1);
 	} else if (gChildProcessId < 0) {
@@ -172,7 +172,6 @@ void tterm_start(TTerm* p, const char* tn, const char* en)
 	}
 	/* parent */
 	tterm_set_utmp(p);
-	util_privilege_drop();
 	signal(SIGCHLD, sigchld);
 	atexit(application_final);
 
