@@ -36,12 +36,14 @@
 #include	<sys/types.h>
 #include	<sys/ioctl.h>
 #include	<sys/kd.h>
+#include	<errno.h>
 
 #include	"util.h"
 #include	"term.h"
 #include	"vterm.h"
 #include	"font.h"
 #include	"csv.h"
+#include	"config.h"
 
 #include	"sequence.h"
 #include	"message.h"
@@ -57,11 +59,23 @@ static void tvterm_esc_designate_other_coding_system(TVterm* p, u_char ch);
 static void tvterm_invoke_gx(TVterm* p, TFontSpec* fs, u_int n);
 static void tvterm_re_invoke_gx(TVterm* p, TFontSpec* fs);
 void tvterm_set_window_size(TVterm* p);
-#if 0
-static void tvterm_esc_set_double_code(TVterm* p, u_char ch);
+
+static void tvterm_switch_to_ISO2022(TVterm *p);
+static int tvterm_is_ISO2022(TVterm *p);
+#ifdef JFB_UTF8
+static void tvterm_switch_to_UTF8(TVterm *p);
+static int tvterm_is_UTF8(TVterm *p);
+#endif
+#ifdef JFB_OTHER_CODING_SYSTEM
+static void tvterm_codingSystem_final(TVterm *p);
+static void tvterm_switch_to_other(TVterm *p, TCodingSystem *other);
+static int tvterm_is_otherCodingSystem(TVterm *p);
+
+#define MAX_OTHER_CODING_SYSTEM		6
+static TCodingSystem gOtherCodingSystem[MAX_OTHER_CODING_SYSTEM];
 #endif
 
-void tvterm_init(TVterm* p, TTerm* pt, u_int hw, u_int hh, const char* en)
+void tvterm_init(TVterm* p, TTerm* pt, u_int hw, u_int hh, TCaps *caps, const char* en)
 {
 	p->term = pt;
 	p->xcap = hw;
@@ -85,51 +99,303 @@ void tvterm_init(TVterm* p, TTerm* pt, u_int hw, u_int hh, const char* en)
 	p->gDefaultIdx[1] = 1;			/* G1 <== JIS X 0208 */
 	p->gDefaultIdx[2] = 0;			/* G2 <== ASCII */
 	p->gDefaultIdx[3] = 0;			/* G3 <== ASCII */
-
+#ifdef JFB_UTF8
+	p->utf8DefaultIdx = 0;
+	p->utf8Idx = 0;
+	p->utf8remain = 0;
+	p->ucs2ch = 0;
+#endif
+#ifdef JFB_OTHER_CODING_SYSTEM
+	p->codingSystem = NULL;
+#endif
+	p->caps = caps;
 	tvterm_set_default_encoding(p, en);
 }
 
-void tvterm_set_default_encoding(TVterm* p, const char* en)
+int
+tvterm_parse_encoding(const char *en, int idx[6])
 {
 	static const char* table[] = {"G0", "G1", "G2", "G3", NULL}; 
-	
-	const char *g;
-	int ig;
-	int i;
 	TCsv farg;
-	
+	const char *g;
+	int i;
+	int ig;
+
 	tcsv_init(&farg, en);
 	if (farg.cap != 6) {
-		print_message("ENCODING : Skipped (BAD FORMAT)\n");
-		goto FINALIZE;
+		tcsv_final(&farg);
+		return 0;
 	}
-	/* GL */
-	g = tcsv_get_token(&farg);
-	ig = util_search_string(g, table);
-	if (ig == -1) {
-		print_message("ENCODING : GL ? (BAD FORMAT)\n");
-		ig = 0;
+	for (i = 0; i < 2; i++) {
+		g = tcsv_get_token(&farg);
+		ig = util_search_string(g, table);
+		if (ig == -1)
+			ig = i;
+		idx[i] = ig;
 	}
-	p->gDefaultL = ig;
-	/* GR */
-	g = tcsv_get_token(&farg);
-	ig = util_search_string(g, table);
-	if (ig == -1) {
-		print_message("ENCODING : GR ? (BAD FORMAT)\n");
-		ig = 1;
-	}
-	p->gDefaultR = ig;
-	/* G0 .. G3 */
 	for (i = 0 ; i < 4 ; i++) {
 		g = tcsv_get_token(&farg);
 		ig = tfont_ary_search_idx(g);
 		if (ig == -1) {
-			print_message("ENCODING : G%d ? (BAD FORMAT)\n", i);
 			ig = 0;
 		}
-		p->gDefaultIdx[i] = ig;
+		idx[2+i] = ig;
 	}
+	tcsv_final(&farg);
+	return 1;
+}
+
+static void
+tvterm_switch_to_ISO2022(TVterm *p)
+{
+#ifdef JFB_OTHER_CODING_SYSTEM
+	tvterm_codingSystem_final(p);
+#endif
+#ifdef JFB_UTF8
+	p->utf8Idx = 0;
+	p->utf8remain = 0;
+	p->ucs2ch = 0;
+#endif
+}
+
+static int
+tvterm_is_ISO2022(TVterm *p)
+{
+#ifdef JFB_OTHER_CODING_SYSTEM
+	if (tvterm_is_otherCodingSystem(p))
+		return 0;
+#endif
+#ifdef JFB_UTF8
+	if (tvterm_is_UTF8(p))
+		return 0;
+#endif
+	return 1;
+}
+
+#ifdef JFB_UTF8
+static int
+tvterm_UTF8index(const char *en)
+{
+	int i;
+	TCsv farg;
+	const char *g;
+	int id = 0;
 	
+	tcsv_init(&farg, en);
+	g = tcsv_get_token(&farg);
+	if (strcmp(g, "UTF-8") != 0)
+		goto FINALIZE;
+	if (farg.cap != 2)
+		goto FINALIZE;
+	g = tcsv_get_token(&farg);
+	for (i = 0; gFont[i].name; i++) {
+		if (strcmp(gFont[i].name, g) == 0) {
+			if (tfont_is_loaded(&(gFont[i]))) {
+				id = i;
+				break;
+			}
+		}
+	}
+FINALIZE:
+	tcsv_final(&farg);
+	return id;
+}
+
+static void
+tvterm_switch_to_UTF8(TVterm *p)
+{
+#ifdef JFB_OTHER_CODING_SYSTEM
+	tvterm_codingSystem_final(p);
+#endif
+	if (p->utf8DefaultIdx == 0) {
+		const char *en = tcaps_find_entry(p->caps,
+						  "encoding.",
+						  "UTF-8");
+		if (en != NULL) {
+			p->utf8DefaultIdx = tvterm_UTF8index(en);
+		}
+	}
+	p->utf8Idx = p->utf8DefaultIdx;
+	p->utf8remain = 0;
+	p->ucs2ch = 0;
+	return;
+}
+
+static int
+tvterm_is_UTF8(TVterm *p) 
+{
+	return (p->utf8Idx != 0);
+}
+
+#endif
+
+#ifdef JFB_OTHER_CODING_SYSTEM
+int
+tvterm_parse_other(const char *en, TCodingSystem* other)
+{
+	TCsv farg;
+	const char *g;
+
+	tcsv_init(&farg, en);
+
+	if (farg.cap != 4) {
+		tcsv_final(&farg);
+		return 0;
+	}
+	g = tcsv_get_token(&farg); /* other */
+	if (other->fromcode)
+		free(other->fromcode);
+	other->fromcode = NULL;
+	if (other->tocode)
+		free(other->tocode);
+	other->tocode = NULL;
+
+	g = tcsv_get_token(&farg);
+	other->fromcode = strdup(g);
+
+	g = tcsv_get_token(&farg);
+	/* g == "iconv" */
+
+	g = tcsv_get_token(&farg);
+	other->tocode = strdup(g);
+	tcsv_final(&farg);
+
+	return 1;
+}
+
+static void
+tvterm_codingSystem_final(TVterm *p)
+{
+	/* XXX restore ISO-2022 state? */
+	if (p->codingSystem) {
+		if (p->codingSystem->cd != (iconv_t)(-1)) {
+			iconv_close(p->codingSystem->cd);
+		}
+		p->codingSystem->cd = (iconv_t)(-1);
+		if (strcmp(p->codingSystem->tocode, "UTF-8") != 0)
+			p->utf8Idx = 0;
+	}
+	p->codingSystem = NULL;
+}
+
+static void 
+tvterm_switch_to_other(TVterm *p, TCodingSystem *other)
+{
+	char *tocode = other->tocode;
+	tvterm_codingSystem_final(p);
+	if (strcmp(other->tocode, "UTF-8") == 0) {
+		tvterm_switch_to_UTF8(p);
+		tocode = "UCS-2BE"; /* XXX */
+	} else {
+		int idx[6];
+		char *en;
+		int i;
+
+		en = tcaps_find_entry(p->caps, "encoding.", other->tocode);
+		if (en == NULL)
+			return;
+		if (tvterm_parse_encoding(en, idx) == 0)
+			return;
+		for (i = 0; i < 4; i++)
+			p->gIdx[i] = idx[2+i];
+		tvterm_invoke_gx(p, &(p->gl), idx[0]);
+		tvterm_invoke_gx(p, &(p->gr), idx[1]);
+		p->tgl = p->gl;
+		p->tgr = p->gr;
+		p->knj1 = 0;
+	}
+	other->cd = iconv_open(tocode, other->fromcode);
+	if (other->cd == (iconv_t)(-1)) {
+		tvterm_codingSystem_final(p);
+		return;
+	}
+	memset(other->inbuf, 0, sizeof(other->inbuf));
+	other->inbuflen = 0;
+	memset(other->outbuf, 0, sizeof(other->outbuf));
+	p->codingSystem = other;
+}
+
+static int
+tvterm_is_otherCodingSystem(TVterm *p)
+{
+	return (p->codingSystem != NULL);
+}
+
+void
+tvterm_setup_otherCodingSystem(TCaps *caps, TCapValue* vp)
+{
+	TCsv farg;
+	const char *fc;
+	const char *ocs;
+	TCodingSystem *other;
+	char *en;
+	int i;
+	for (i = 0; vp; vp = vp->next) {
+		if (i >= MAX_OTHER_CODING_SYSTEM)
+			return;
+		tcsv_init(&farg, vp->value);
+		if (farg.cap != 2) {
+			print_message("other-coding-system: skipped\n");
+			goto FINALIZE;
+		}
+		fc = tcsv_get_token(&farg);
+		ocs = tcsv_get_token(&farg);
+		other = &gOtherCodingSystem[i];
+		other->fch = fc[0];
+		en = tcaps_find_entry(caps, "encoding.", ocs);
+		if (en == NULL)
+			goto FINALIZE;
+		if (tvterm_parse_other(en, other))
+			i++;
+	FINALIZE:
+		tcsv_final(&farg);
+	}
+}
+#endif
+
+
+void tvterm_set_default_encoding(TVterm* p, const char* en)
+{
+	const char *g;
+	int i;
+	int idx[6];
+	TCsv farg;
+	int ncap;
+
+	tcsv_init(&farg, en);
+	ncap = farg.cap;
+	/* first token */
+	g = tcsv_get_token(&farg);
+#ifdef JFB_OTHER_CODING_SYSTEM
+	if (strcmp(g, "other") == 0) {
+		/* other,<coding-system>,iconv,<std-coding-system> */
+		static TCodingSystem other;
+		if (tvterm_parse_other(en, &other) == 0)
+			goto FINALIZE;
+		tvterm_switch_to_other(p, &other);
+		goto FINALIZE;
+	}
+#endif
+#ifdef JFB_UTF8
+	if (strcmp(g, "UTF-8") == 0) {
+		/* UTF-8,<fontsetname> */
+		p->utf8DefaultIdx = tvterm_UTF8index(en);
+		tvterm_switch_to_UTF8(p);
+		goto FINALIZE;
+	}
+#endif
+	/* ISO-2022 */
+	if (tvterm_parse_encoding(en, idx) == 0) {
+		goto FINALIZE;
+	}
+	/* GL */
+	p->gDefaultL = idx[0];
+	/* GR */
+	p->gDefaultR = idx[1];
+	/* G0 .. G3 */
+	for (i = 0 ; i < 4 ; i++) {
+		p->gDefaultIdx[i] = idx[2+i];
+	}
 FINALIZE:	
 	tcsv_final(&farg);
 }
@@ -145,10 +411,13 @@ void tvterm_set_default_invoke_and_designate(TVterm* p)
 	p->tgl = p->gl;				/* Next char's GL <== GL */
 	p->tgr = p->gr;				/* Next char's GR <== GR */
 	p->knj1 = 0;
+#ifdef JFB_UTF8
+	p->utf8Idx = p->utf8DefaultIdx;
+	p->utf8remain = 0;
+	p->ucs2ch = 0;
+#endif
 }
 
-
-
 void tvterm_start(TVterm* p)
 {
 	p->pen.x = 0;
@@ -279,7 +548,7 @@ static inline void SET_WARP_FLAG_IF_NEEDED(TVterm* p)
 }
 
 
-static int tvterm_put_normal_char(TVterm* p, u_char ch, const char *buff)
+static int tvterm_put_normal_char(TVterm* p, u_char ch)
 {
 	if (p->pen.x == p->xmax) {
 		p->wrap = TRUE;
@@ -338,6 +607,41 @@ static int tvterm_put_normal_char(TVterm* p, u_char ch, const char *buff)
 	return 0;
 }
 
+#ifdef JFB_UTF8
+static int tvterm_put_uchar(TVterm* p, u_int ch)
+{
+	TFont *pf = &gFont[p->utf8Idx];
+	u_int w;
+	if (p->pen.x == p->xmax) {
+		p->wrap = TRUE;
+		p->pen.x --;
+	}
+	if (p->wrap) {
+		p->pen.x -= p->xmax -1;
+		if (p->pen.y == p->ymax -1) {
+			p->scroll++;
+		} else {
+			p->pen.y ++;
+		}
+		p->wrap = FALSE;
+		return -1;
+	}
+	pf->conv(pf, ch, &w);
+	if (pf->width == w) {
+		INSERT_N_CHARS_IF_NEEDED(p, 1);
+		tvterm_uput1(p, p->utf8Idx, ch);
+		p->pen.x ++;
+	} else {
+		INSERT_N_CHARS_IF_NEEDED(p, 2);
+		tvterm_uput2(p, p->utf8Idx, ch);
+		p->pen.x += 2;
+	}
+	p->utf8remain = 0;
+	p->ucs2ch = 0;
+	return 0;
+}
+#endif
+
 int tvterm_iso_C0_set(TVterm* p, u_char ch)
 {
 	switch(ch) {
@@ -357,7 +661,10 @@ int tvterm_iso_C0_set(TVterm* p, u_char ch)
 		/* fail into next case */
 	case ISO_VT:
 	case ISO_FF:
-		tvterm_set_default_invoke_and_designate(p);
+#ifdef JFB_OTHER_CODING_SYSTEM
+		if (!tvterm_is_otherCodingSystem(p))
+#endif
+			tvterm_set_default_invoke_and_designate(p);
 		/* fail into next case */
 	case ISO_LF:
 		p->wrap = FALSE;
@@ -377,10 +684,14 @@ int tvterm_iso_C0_set(TVterm* p, u_char ch)
 		p->escGn = 0;
 		return 1;
 	case ISO_LS1:
+		if (!tvterm_is_ISO2022(p))
+			break;
 		tvterm_invoke_gx(p, &(p->gl), 1); /* GL <== G1 */
 		p->tgl = p->gl;
 		return 1;
 	case ISO_LS0:
+		if (!tvterm_is_ISO2022(p))
+			break;
 		tvterm_invoke_gx(p, &(p->gl), 0); /* GL <== G0 */
 		p->tgl = p->gl;
 		return 1;
@@ -392,19 +703,138 @@ int tvterm_iso_C0_set(TVterm* p, u_char ch)
 
 int tvterm_iso_C1_set(TVterm* p, u_char ch)
 {
-    switch (ch) {
-    case ISO_SS2:	/* single shift 2 */
-	tvterm_invoke_gx(p, &(p->tgr), 2); /* GR <== G2 */
-	return 1;
-    case ISO_SS3:	/* single shift 3 */
-	tvterm_invoke_gx(p, &(p->tgr), 3); /* GR <== G3 */
-	return 1;
-    default:
-	break;
-    }
-    return 0;
+	switch (ch) {
+	case ISO_SS2:	/* single shift 2 */
+		tvterm_invoke_gx(p, &(p->tgr), 2); /* GR <== G2 */
+		return 1;
+	case ISO_SS3:	/* single shift 3 */
+		tvterm_invoke_gx(p, &(p->tgr), 3); /* GR <== G3 */
+		return 1;
+	default:
+		break;
+	}
+	return 0;
 }
 
+#ifdef JFB_UTF8
+#define UTF8_CHECK_START(p) { \
+	if ((p)->utf8remain != 0) { \
+		(p)->ucs2ch = 0; \
+		(p)->utf8remain = 0; \
+	} \
+}
+
+int
+tvterm_put_utf8_char(TVterm *p, u_char ch)
+{
+	int rev;
+	if (ch < 0x7F) {
+		UTF8_CHECK_START(p);
+		p->ucs2ch = ch;
+	} else if ((ch & 0xC0) == 0x80) {
+		if (p->utf8remain == 0) {
+			/* illegal UTF-8 sequence? */
+			p->ucs2ch = ch;
+		} else {
+			p->ucs2ch <<= 6;
+			p->ucs2ch |= (ch & 0x3F);
+			p->utf8remain--;
+			if (p->utf8remain > 0) {
+				return p->utf8remain;
+			}
+		}
+	} else if ((ch & 0xE0) == 0xC0) {
+		UTF8_CHECK_START(p);
+		p->utf8remain = 1;
+		p->ucs2ch = (ch & 0x1F);
+		return p->utf8remain;
+	} else if ((ch & 0xF0) == 0xE0) {
+		UTF8_CHECK_START(p);
+		p->utf8remain = 2;
+		p->ucs2ch = (ch & 0x0F);
+		return p->utf8remain;
+	} else {
+		/* XXX: support UCS2 only */
+	}
+	rev = tvterm_put_uchar(p, p->ucs2ch);
+	if (rev < 0) {
+		p->ucs2ch >>= 6;
+		if (p->ucs2ch)
+			p->utf8remain ++;
+	}
+	return rev;
+}
+#endif
+
+#ifdef JFB_OTHER_CODING_SYSTEM
+int 
+tvterm_put_other_char(TVterm *p, u_char ch)
+{
+	int rev;
+	char *inbuf;
+	size_t inbuflen;
+	char *outbuf;
+	size_t outbuflen;
+	size_t s;
+	int i;
+
+	if (p->codingSystem->inbuflen >= sizeof(p->codingSystem->inbuf)-1) {
+		/* XXX: too long? */
+		p->codingSystem->inbuflen = 0;
+	}
+	p->codingSystem->inbuf[p->codingSystem->inbuflen] = ch;
+	p->codingSystem->inbuflen++;
+
+	inbuf = p->codingSystem->inbuf;
+	inbuflen = p->codingSystem->inbuflen;
+	outbuf = p->codingSystem->outbuf;
+	outbuflen = sizeof(p->codingSystem->outbuf);
+
+	s = iconv(p->codingSystem->cd, &inbuf, &inbuflen, &outbuf, &outbuflen);
+	if (s == (size_t)(-1)) {
+		switch (errno) {
+		case E2BIG:
+			p->codingSystem->inbuflen = 0;
+			return 0;
+		case EILSEQ: /* illegal sequence */
+			p->codingSystem->inbuflen = 0;
+			return 0;
+		case EINVAL: /* incomplete multibyte */
+			return 0;
+		}
+	}
+	/* output bytes */
+	s = sizeof(p->codingSystem->outbuf) - outbuflen;
+
+	if (strcmp(p->codingSystem->tocode, "UTF-8") == 0) {
+		p->ucs2ch = 0;
+		for (i = 0; i < s; i++) {
+			p->ucs2ch <<= 8;
+			p->ucs2ch |= (p->codingSystem->outbuf[i] & 0x0ff);
+		}
+
+		rev = tvterm_put_uchar(p, p->ucs2ch);
+		if (rev < 0) {
+			p->codingSystem->inbuflen--;
+			return rev;
+		}
+		p->codingSystem->inbuflen = 0;
+		return rev;
+	} else {
+		for (i = 0; i < s; i++) {
+			u_char c = p->codingSystem->outbuf[i];
+			rev = tvterm_put_normal_char(p, c);
+			if (rev < 0) {
+				p->codingSystem->inbuflen--;
+				return rev;
+			}
+		}
+		p->codingSystem->inbuflen = 0;
+		return 0;
+	}
+}
+#endif
+
 void tvterm_emulate(TVterm* p, const char *buff, int nchars)
 {
 	u_char	ch;
@@ -422,6 +852,26 @@ void tvterm_emulate(TVterm* p, const char *buff, int nchars)
 			if (cn) {
 				continue;
 			}
+#ifdef JFB_OTHER_CODING_SYSTEM
+		} else if (tvterm_is_otherCodingSystem(p)) {
+			rev = tvterm_put_other_char(p, ch);
+			if (rev >= 0) {
+				continue;
+			} else if (rev < 0) {
+				nchars -= rev;
+				buff += rev;
+			}
+#endif
+#ifdef JFB_UTF8
+		} else if (tvterm_is_UTF8(p)) {
+			rev = tvterm_put_utf8_char(p, ch);
+			if (rev >= 0) {
+				continue;
+			} else if (rev < 0) {
+				nchars -= rev;
+				buff += rev;
+			}
+#endif
 		} else if ((0x7F < ch) && (ch < 0xA0)) {
 			cn = tvterm_iso_C1_set(p, ch);
 			if (cn) {
@@ -430,7 +880,7 @@ void tvterm_emulate(TVterm* p, const char *buff, int nchars)
 		} else if (ch == ISO_DEL) {
 			/* nothing to do. */
 		} else {
-			rev = tvterm_put_normal_char(p, ch, buff);
+			rev = tvterm_put_normal_char(p, ch);
 			if (rev == 0) {
 				continue;
 			} else if (rev < 0) {
@@ -538,22 +988,32 @@ static void tvterm_esc_start(TVterm* p, u_char ch)
 		p->wrap = FALSE;
 		break;
 	case ISO_LS2:
+		if (!tvterm_is_ISO2022(p))
+			break;
 		tvterm_invoke_gx(p, &(p->gl), 2); /* GL <= G2 */
 		p->tgl = p->gl;
 		break;
 	case ISO_LS3:
+		if (!tvterm_is_ISO2022(p))
+			break;
 		tvterm_invoke_gx(p, &(p->gl), 3); /* GL <= G3 */
 		p->tgl = p->gl;
 		break;
 	case ISO_LS3R:
+		if (!tvterm_is_ISO2022(p))
+			break;
 		tvterm_invoke_gx(p, &(p->gr), 3); /* GR <= G3 */
 		p->tgr = p->gr;
 		break;
 	case ISO_LS2R:
+		if (!tvterm_is_ISO2022(p))
+			break;
 		tvterm_invoke_gx(p, &(p->gr), 2); /* GR <= G2 */
 		p->tgr = p->gr;
 		break;
 	case ISO_LS1R:
+		if (!tvterm_is_ISO2022(p))
+			break;
 		tvterm_invoke_gx(p, &(p->gr), 1); /* GR <= G1 */
 		p->tgr = p->gr;
 		break;
@@ -871,6 +1331,10 @@ static void tvterm_re_invoke_gx(TVterm* p, TFontSpec* fs)
 static void tvterm_esc_designate_font(TVterm* p, u_char ch)
 {
 	int i;
+	if (!tvterm_is_ISO2022(p)) {
+		p->esc = NULL;
+		return;
+	}
 
 	for (i = 0 ; gFont[i].fsignature ; i++) {
 		if (gFont[i].fsignature  == (ch | p->escSignature)) {
@@ -889,11 +1353,26 @@ static void tvterm_esc_designate_other_coding_system(TVterm* p, u_char ch)
 	switch (ch) {
 	case 0x40:
 		/* reset ISO-2022 */
+		tvterm_switch_to_ISO2022(p);
 		break;
-#if 0
-	case 0x47:
-		/* designate and invoke UTF-8 coding */
+#ifdef JFB_UTF8
+	case 0x47: /* XXX: designate and invoke UTF-8 coding */
+		tvterm_switch_to_UTF8(p);
 		break;
+#endif
+#ifdef JFB_OTHER_CODING_SYSTEM
+	default:
+	{
+		int i;
+		/* XXX: other coding system? */
+		for (i = 0; i < MAX_OTHER_CODING_SYSTEM; i++) {
+			if (ch == gOtherCodingSystem[i].fch) {
+				tvterm_switch_to_other(p, 
+						       &gOtherCodingSystem[i]);
+				break;
+			}
+		}
+	}
 #endif
 	}
 	p->esc = NULL;
@@ -908,56 +1387,85 @@ static void tvterm_esc_traditional_multibyte_fix(TVterm* p, u_char ch)
 	}
 }
 
-#if 0
-static void tvterm_esc_set_double_code(TVterm* p, u_char ch)
+
+
+void tvterm_show_sequence(FILE *tf, TCaps *cap, const char *en)
 {
-	switch(ch) {
-	case ISO_GZD4:
-		p->escSignature = TFONT_FT_94CHAR | TFONT_FT_DOUBLE;
-		p->escGn = 0;
-		p->esc = tvterm_esc_designate_font;
-		break;
-	case ISO_G1D4:
-		p->escSignature = TFONT_FT_94CHAR | TFONT_FT_DOUBLE;
-		p->escGn = 1;
-		p->esc = tvterm_esc_designate_font;
-		break;
-	case ISO_G2D4:
-		p->escSignature = TFONT_FT_94CHAR | TFONT_FT_DOUBLE;
-		p->escGn = 2;
-		p->esc = tvterm_esc_designate_font;
-		break;
-	case ISO_G3D4:
-		p->escSignature = TFONT_FT_94CHAR | TFONT_FT_DOUBLE;
-		p->escGn = 3;
-		p->esc = tvterm_esc_designate_font;
-		break;
-	case MULE__GZD6:
-		p->escSignature = TFONT_FT_96CHAR | TFONT_FT_DOUBLE;
-		p->escGn = 0;
-		p->esc = tvterm_esc_designate_font;
-		break;
-	case ISO_G1D6:
-		p->escSignature = TFONT_FT_96CHAR | TFONT_FT_DOUBLE;
-		p->escGn = 1;
-		p->esc = tvterm_esc_designate_font;
-		break;
-	case ISO_G2D6:
-		p->escSignature = TFONT_FT_96CHAR | TFONT_FT_DOUBLE;
-		p->escGn = 2;
-		p->esc = tvterm_esc_designate_font;
-		break;
-	case ISO_G3D6:
-		p->escSignature = TFONT_FT_96CHAR | TFONT_FT_DOUBLE;
-		p->escGn = 3;
-		p->esc = tvterm_esc_designate_font;
-		break;
-	default:
-		tvterm_esc_designate_font(p, ch);
-		p->esc = NULL;
-		break;
+	static const char *invokeL[] = {"\017", "\016", "\033n", "\033o"};
+	static const char *invokeR[] = {"", "\033~", "\033}", "\033|"};
+	static const char csr4[] = {ISO_GZD4, ISO_G1D4, ISO_G2D4, ISO_G3D4};
+	static const char csr6[] = {MULE__GZD6, ISO_G1D6, ISO_G2D6, ISO_G3D6};
+	TCsv farg;
+	int i;
+	const char *g;
+	int idx[6];
+
+	tcsv_init(&farg, en);
+	g = tcsv_get_token(&farg);
+#ifdef JFB_OTHER_CODING_SYSTEM
+	if (strcmp(g, "other") == 0) {
+		TCodingSystem tcs;
+		memset(&tcs, 0, sizeof(tcs));
+		if (tvterm_parse_other(en, &tcs)) {
+			TCapability* fother;
+			TCapValue* vp;
+			fother = tcaps_find(cap, "other-coding-system");
+			if (!fother || !(fother->values))
+				goto end;
+			for (vp = fother->values; vp ; vp = vp->next) {
+				TCsv fo;
+				const char *fc;
+				const char *ocs;
+				tcsv_init(&fo, vp->value);
+				if (fo.cap != 2) {
+					tcsv_final(&fo);
+					continue;
+				}
+				fc = tcsv_get_token(&fo);
+				ocs = tcsv_get_token(&fo);
+				if (strcmp(tcs.fromcode, ocs) == 0) {
+					fprintf(tf, "%s%c", "\033%", fc[0]);
+					tcsv_final(&fo);
+					break;
+				}
+				tcsv_final(&fo);
+			}
+		}
+		goto end;
 	}
-}
 #endif
-
+#ifdef JFB_UTF8
+	if (strcmp(g, "UTF-8") == 0) {
+		fprintf(tf, "%s", "\033%G");
+		goto end;
+	}
+#endif
+	/* ISO-2022 */
+	fprintf(tf, "%s", "\033%@");
+	if (tvterm_parse_encoding(en, idx) == 0) {
+		print_warn("ENCODING : BAD FORMAT : %s\n", en);
+		goto end;
+	}
+	/* GL */
+	fprintf(tf, "%s", invokeL[idx[0]]);
+	/* GR */
+	fprintf(tf, "%s", invokeR[idx[1]]);
+	/* G0 .. G3 */
+	for (i = 0; i < 4; i++) {
+		TFont *f;
+		u_int n;
+		char c;
+		u_int f1, f2;
+		f = &gFont[idx[2+i]];
+		n = f->fsignature;
+		c = n & 255;
+		f1 = n & TFONT_FT_DOUBLE;
+		f2 = n & TFONT_FT_96CHAR;
+		fprintf(tf, "\033%s%c%c", (f1 ? "$" : ""), 
+			(f2 ? csr6[i] : csr4[i]), c);
+	}
+end:
+	tcsv_final(&farg);
+	return;
+}
 
