@@ -36,6 +36,11 @@
 #include	"mytypes.h"
 #include	"pen.h"
 #include	"font.h"
+#include	"config.h"
+
+#ifdef JFB_OTHER_CODING_SYSTEM
+#include <iconv.h>
+#endif
 
 #define	LEN_REPORT	9
 
@@ -47,6 +52,22 @@ typedef struct Raw_TFontSpec {
 	FONTSET_HALF half;	/* ʸG0,G1 ΤɤȤäƤ뤫 */
 } TFontSpec;
 
+#ifdef JFB_OTHER_CODING_SYSTEM
+typedef struct Raw_TCodingSystem {
+	u_int fch;
+	/* iconv state */
+	char *fromcode;
+	char *tocode;
+	iconv_t cd;
+#define MAX_MULTIBYTE_CHARLEN	6
+	char inbuf[MAX_MULTIBYTE_CHARLEN];
+	int inbuflen;
+	char outbuf[MAX_MULTIBYTE_CHARLEN];
+} TCodingSystem;
+
+void tvterm_setup_otherCodingSystem(TCaps *caps, TCapValue* vp);
+#endif
+
 typedef struct Raw_TCursor {
 	u_int x;
 	u_int y;
@@ -71,20 +92,6 @@ typedef struct Raw_TVterm {
 	TPen* savedPenSL;		/* ơ饤 */
 	int scroll;			/* Կ */
 	/* -- */
-#if 0
-	u_char sFontCode;		/* 1 byte code եֹ */
-	u_char dFontCode;		/* 2 byte code եֹ */
-	u_char secondCode;		/* ⤦Ĥδ */
-	enum {
-		CS_LEFT,
-		CS_RIGHT,
-		CS_GRAPH,
-		CS_DBCS
-	} trans, g0, g1;
-	/*  */;
-	u_char sLanguage;		/* 1 byte code ֹ */
-	u_char dLanguage;		/* 2 byte code ֹ */
-#endif
 
 	u_char knj1;			/* first byte of 2 byte code */
 	FONTSET_HALF knj1h;
@@ -108,6 +115,16 @@ typedef struct Raw_TVterm {
 		SL_ENTER,
 		SL_LEAVE
 	} sl;
+#ifdef JFB_UTF8
+	u_int utf8DefaultIdx;
+	u_int utf8Idx;
+	u_int utf8remain;
+	u_int ucs2ch;
+#endif
+#ifdef JFB_OTHER_CODING_SYSTEM
+	TCodingSystem *codingSystem;
+#endif
+	TCaps *caps;
 
 	TBool soft;
 	TBool wrap;
@@ -145,18 +162,24 @@ void tvterm_text_clear_eol(TVterm* p, u_char mode);
 void tvterm_text_clear_eos(TVterm* p, u_char mode);
 void tvterm_wput(TVterm* p, u_int idx, u_char ch1, u_char ch2);
 void tvterm_sput(TVterm* p, u_int idx, u_char ch);
+#ifdef JFB_UTF8
+void tvterm_uput1(TVterm* p, u_int idx, u_int ch);
+void tvterm_uput2(TVterm* p, u_int idx, u_int ch);
+#endif
 void tvterm_text_clear_all(TVterm* p);
 
 void tvterm_emulate(TVterm* p, const char *buff, int nchars);
 void tvterm_refresh(TVterm* p);
 
-void tvterm_init(TVterm* p, struct Raw_TTerm* tp, u_int hx, u_int hy, const char* en);
+void tvterm_init(TVterm* p, struct Raw_TTerm* tp, u_int hx, u_int hy, TCaps *caps, const char* en);
 void tvterm_start(TVterm* p);
 void tvterm_final(TVterm* p);
 
 void tvterm_unregister_signal(void);
 void tvterm_register_signal(TVterm* p);
 
+void tvterm_show_sequence(FILE *fp, TCaps *cap, const char *en);
+
 
 /*
 
