@@ -379,6 +379,7 @@ void tvterm_refresh(TVterm* p)
 	u_char	fc, bc, fg;
 	TFont* pf;
 	const u_char* glyph;
+	u_int  w, gw;
 
 	p->busy = TRUE;
 	if (!p->active) {
@@ -396,6 +397,7 @@ void tvterm_refresh(TVterm* p)
 
 	for (y = 0; y < p->ycap; y ++) {
 		for (x = 0; x < p->xcap; x ++) {
+			w = 1;
 			i = tvterm_coord_to_index(p, x, y);
 			fg = p->flag[i];
 			if (fg & CLEAN_S) {
@@ -413,6 +415,7 @@ void tvterm_refresh(TVterm* p)
 				pf = &(gFont[lang]);
 				i++;
 				p->flag[i] |= CLEAN_S;
+				w = 2;
 #if 0
 				if (p->ins) {
 					tvterm_insert_n_chars(p, 2);
@@ -426,14 +429,15 @@ void tvterm_refresh(TVterm* p)
 				}
 #endif
 			}
-			glyph = pf->conv(pf, chlw);
+			/* XXX: multiwidth support (unifont) */
+			glyph = pf->conv(pf, chlw, &gw);
 			gFramebuffer.cap.fill(&gFramebuffer,
 				gFontsWidth * x, gFontsHeight * y,
-				gFontsWidth + (ch2 ? gFontsWidth : 0),
+				gFontsWidth * w,
 				gFontsHeight, bc);
 			gFramebuffer.cap.overlay(&gFramebuffer,
 				gFontsWidth * x, gFontsHeight * y,
-				glyph, pf->width, pf->height, pf->bytew, fc);
+				glyph, gw, pf->height, pf->bytew, fc);
 		}
 	}
 	tvterm_set_cursor_wide(p, IsKanji(p,p->pen.x,p->pen.y));
@@ -537,6 +541,28 @@ void tvterm_sput(TVterm* p, u_int idx, u_char ch)
 	p->flag[a] = LATCH_S;
 }
 
+#ifdef JFB_UTF8
+void tvterm_uput1(TVterm* p, u_int idx, u_int ch)
+{
+	u_int a = tvterm_coord_to_index(p, p->pen.x, p->pen.y);
+
+	p->attr[a] = p->pen.fcol | (p->pen.bcol << 4);
+	p->text[a] = (idx << 24) | ch;
+	p->flag[a] = LATCH_S;
+}
+
+void tvterm_uput2(TVterm* p, u_int idx, u_int ch)
+{
+	u_int a= tvterm_coord_to_index(p, p->pen.x, p->pen.y);
+
+	p->attr[a] = p->pen.fcol | (p->pen.bcol << 4);
+	p->text[a] = (idx << 24) | ch;
+	p->text[a+1] = 0;
+	p->flag[a] = LATCH_1;
+	p->flag[a+1] = LATCH_2;
+}
+#endif
+
 /**
 	Զ [0, TVterm::ymax) 򥯥ꥢ롣
 **/
